@@ -23,7 +23,7 @@
 
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/eval.c,v 1.116 2012/04/06 14:07:52 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/eval.c,v 1.118 2012/05/04 22:05:00 tg Exp $");
 
 /*
  * string expansion
@@ -392,7 +392,7 @@ expand(const char *cp,	/* input word */
 						h = seed;
 						NZATUpdateString(h,
 						    str_val(st->var));
-						NZATFinish(h);
+						NZAATFinish(h);
 						x.str = shf_smprintf("%08X",
 						    (unsigned int)h);
 						break;
@@ -1437,7 +1437,7 @@ globit(XString *xs,	/* dest string */
 			struct stat lstatb, statb;
 			int stat_done = 0;	 /* -1: failed, 1 ok */
 
-			if (lstat(Xstring(*xs, xp), &lstatb) < 0)
+			if (mksh_lstat(Xstring(*xs, xp), &lstatb) < 0)
 				return;
 			/*
 			 * special case for systems which strip trailing
