@@ -5,7 +5,7 @@
 
 /*-
  * Copyright (c) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
- *		 2010, 2011
+ *		 2010, 2011, 2012
  *	Thorsten Glaser <tg@mirbsd.org>
  *
  * Provided that these terms and disclaimer and all copyright notices
@@ -38,7 +38,7 @@
 #endif
 #endif
 
-__RCSID("$MirOS: src/bin/mksh/funcs.c,v 1.204 2011/12/09 20:40:25 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/funcs.c,v 1.219 2012/05/09 23:20:56 tg Exp $");
 
 #if HAVE_KILLPG
 /*
@@ -60,9 +60,16 @@ __RCSID("$MirOS: src/bin/mksh/funcs.c,v 1.204 2011/12/09 20:40:25 tg Exp $");
 #define c_ulimit	c_true
 #endif
 
-#if defined(ANDROID)
-static int c_android_lsmod(const char **);
-#endif
+/* getn() that prints error */
+static int
+bi_getn(const char *as, int *ai)
+{
+	int rv;
+
+	if (!(rv = getn(as, ai)))
+		bi_errorf("%s: %s", as, "bad number");
+	return (rv);
+}
 
 static int
 c_true(const char **wp MKSH_A_UNUSED)
@@ -104,7 +111,7 @@ const struct builtin mkshbuiltins[] = {
 	{"+true", c_true},
 	{"ulimit", c_ulimit},
 	{"+umask", c_umask},
-	{"*=unset", c_unset},
+	{Tsgunset, c_unset},
 	/* no =: AT&T manual wrong */
 	{Tpalias, c_alias},
 	{"+cd", c_cd},
@@ -112,7 +119,7 @@ const struct builtin mkshbuiltins[] = {
 	{"chdir", c_cd},
 	{"+command", c_command},
 	{"echo", c_print},
-	{"*=export", c_typeset},
+	{Tsgexport, c_typeset},
 	{"+fc", c_fc},
 	{"+getopts", c_getopts},
 	{"=global", c_typeset},
@@ -124,7 +131,7 @@ const struct builtin mkshbuiltins[] = {
 	{"printf", c_printf},
 #endif
 	{"pwd", c_pwd},
-	{"*=readonly", c_typeset},
+	{Tsgreadonly, c_typeset},
 	{T_typeset, c_typeset},
 	{Tpunalias, c_unalias},
 	{"whence", c_whence},
@@ -132,7 +139,9 @@ const struct builtin mkshbuiltins[] = {
 	{"+bg", c_fgbg},
 	{"+fg", c_fgbg},
 #endif
+#ifndef MKSH_NO_CMDLINE_EDITING
 	{"bind", c_bind},
+#endif
 	{"cat", c_cat},
 #if HAVE_MKNOD
 	{"mknod", c_mknod},
@@ -145,9 +154,6 @@ const struct builtin mkshbuiltins[] = {
 #ifdef __MirBSD__
 	/* alias to "true" for historical reasons */
 	{"domainname", c_true},
-#endif
-#if defined(ANDROID)
-	{"lsmod", c_android_lsmod},
 #endif
 	{NULL, (int (*)(const char **))NULL}
 };
@@ -547,7 +553,7 @@ c_whence(const char **wp)
 				    Talias);
 			if (!iam_whence && !vflag)
 				shprintf("%s %s=", Talias, id);
-			print_value_quoted(tp->val.s);
+			print_value_quoted(shl_stdout, tp->val.s);
 			break;
 		case CFUNC:
 			if (vflag) {
@@ -610,13 +616,16 @@ c_command(const char **wp)
 }
 
 /* typeset, global, export, and readonly */
+static void c_typeset_vardump(struct tbl *, uint32_t, int, bool, bool);
+static void c_typeset_vardump_recursive(struct block *, uint32_t, int, bool,
+    bool);
 int
 c_typeset(const char **wp)
 {
-	struct block *l;
 	struct tbl *vp, **p;
 	uint32_t fset = 0, fclr = 0, flag;
-	int thing = 0, field = 0, base = 0, optc;
+	int thing = 0, field = 0, base = 0, i;
+	struct block *l;
 	const char *opts;
 	const char *fieldstr = NULL, *basestr = NULL;
 	bool localv = false, func = false, pflag = false, istset = true;
@@ -659,9 +668,9 @@ c_typeset(const char **wp)
 	 * Here, the number must follow the RLZi option, but is optional
 	 * (see the # kludge in ksh_getopt()).
 	 */
-	while ((optc = ksh_getopt(wp, &builtin_opt, opts)) != -1) {
+	while ((i = ksh_getopt(wp, &builtin_opt, opts)) != -1) {
 		flag = 0;
-		switch (optc) {
+		switch (i) {
 		case 'L':
 			flag = LJUST;
 			fieldstr = builtin_opt.optarg;
@@ -787,8 +796,10 @@ c_typeset(const char **wp)
 	}
 
 	/* set variables and attributes */
-	if (wp[builtin_opt.optind]) {
-		int i, rv = 0;
+	if (wp[builtin_opt.optind] &&
+	    /* not "typeset -p varname" */
+	    !(!func && pflag && !(fset | fclr))) {
+		int rv = 0;
 		struct tbl *f;
 
 		if (localv && !func)
@@ -820,6 +831,7 @@ c_typeset(const char **wp)
 		return (rv);
 	}
 
+	set_refflag = SRF_NOP;
 	/* list variables and attributes */
 
 	/* no difference at this point.. */
@@ -838,139 +850,135 @@ c_typeset(const char **wp)
 				shf_putc('\n', shl_stdout);
 			}
 		}
-	} else {
-		for (l = e->loc; l; l = l->next) {
-			for (p = ktsort(&l->vars); (vp = *p++); ) {
-				struct tbl *tvp;
-				bool any_set = false;
-				/*
-				 * See if the parameter is set (for arrays, if any
-				 * element is set).
-				 */
-				for (tvp = vp; tvp; tvp = tvp->u.array)
-					if (tvp->flag & ISSET) {
-						any_set = true;
-						break;
-					}
-
-				/*
-				 * Check attributes - note that all array elements
-				 * have (should have?) the same attributes, so checking
-				 * the first is sufficient.
-				 *
-				 * Report an unset param only if the user has
-				 * explicitly given it some attribute (like export);
-				 * otherwise, after "echo $FOO", we would report FOO...
-				 */
-				if (!any_set && !(vp->flag & USERATTRIB))
-					continue;
-				if (flag && (vp->flag & flag) == 0)
-					continue;
-				for (; vp; vp = vp->u.array) {
-					/*
-					 * Ignore array elements that aren't
-					 * set unless there are no set elements,
-					 * in which case the first is reported on
-					 */
-					if ((vp->flag&ARRAY) && any_set &&
-					    !(vp->flag & ISSET))
-						continue;
-					/* no arguments */
-					if (thing == 0 && flag == 0) {
-						/*
-						 * AT&T ksh prints things
-						 * like export, integer,
-						 * leftadj, zerofill, etc.,
-						 * but POSIX says must
-						 * be suitable for re-entry...
-						 */
-						shf_puts("typeset ", shl_stdout);
-						if (((vp->flag&(ARRAY|ASSOC))==ASSOC))
-							shprintf("%s ", "-n");
-						if ((vp->flag&INTEGER))
-							shprintf("%s ", "-i");
-						if ((vp->flag&EXPORT))
-							shprintf("%s ", "-x");
-						if ((vp->flag&RDONLY))
-							shprintf("%s ", "-r");
-						if ((vp->flag&TRACE))
-							shprintf("%s ", "-t");
-						if ((vp->flag&LJUST))
-							shprintf("-L%d ", vp->u2.field);
-						if ((vp->flag&RJUST))
-							shprintf("-R%d ", vp->u2.field);
-						if ((vp->flag&ZEROFIL))
-							shprintf("%s ", "-Z");
-						if ((vp->flag&LCASEV))
-							shprintf("%s ", "-l");
-						if ((vp->flag&UCASEV_AL))
-							shprintf("%s ", "-u");
-						if ((vp->flag&INT_U))
-							shprintf("%s ", "-U");
-						shf_puts(vp->name, shl_stdout);
-						if (pflag) {
-							char *s = str_val(vp);
-
-							shf_putc('=', shl_stdout);
-							/*
-							 * AT&T ksh can't have
-							 * justified integers...
-							 */
-							if ((vp->flag &
-							    (INTEGER|LJUST|RJUST)) ==
-							    INTEGER)
-								shf_puts(s, shl_stdout);
-							else
-								print_value_quoted(s);
-						}
-						shf_putc('\n', shl_stdout);
-						if (vp->flag & ARRAY)
-							break;
-					} else {
-						if (pflag)
-							shf_puts(istset ?
-							    "typeset " :
-							    (flag & EXPORT) ?
-							    "export " :
-							    "readonly ",
-							    shl_stdout);
-						if ((vp->flag&ARRAY) && any_set)
-							shprintf("%s[%lu]",
-							    vp->name,
-							    arrayindex(vp));
-						else
-							shf_puts(vp->name, shl_stdout);
-						if (thing == '-' && (vp->flag&ISSET)) {
-							char *s = str_val(vp);
-
-							shf_putc('=', shl_stdout);
-							/*
-							 * AT&T ksh can't have
-							 * justified integers...
-							 */
-							if ((vp->flag &
-							    (INTEGER|LJUST|RJUST)) ==
-							    INTEGER)
-								shf_puts(s, shl_stdout);
-							else
-								print_value_quoted(s);
-						}
-						shf_putc('\n', shl_stdout);
-					}
-					/*
-					 * Only report first 'element' of an array with
-					 * no set elements.
-					 */
-					if (!any_set)
-						break;
-				}
-			}
+	} else if (wp[builtin_opt.optind]) {
+		for (i = builtin_opt.optind; wp[i]; i++) {
+			varsearch(e->loc, &vp, wp[i], hash(wp[i]));
+			if (!vp)
+				continue;
+			c_typeset_vardump(vp, flag, thing, pflag, istset);
 		}
-	}
-	set_refflag = SRF_NOP;
+	} else
+		c_typeset_vardump_recursive(e->loc, flag, thing, pflag, istset);
 	return (0);
 }
 
+static void
+c_typeset_vardump_recursive(struct block *l, uint32_t flag, int thing,
+    bool pflag, bool istset)
+{
+	struct tbl **blockvars, *vp;
+
+	if (l->next)
+		c_typeset_vardump_recursive(l->next, flag, thing, pflag, istset);
+	blockvars = ktsort(&l->vars);
+	while ((vp = *blockvars++))
+		c_typeset_vardump(vp, flag, thing, pflag, istset);
+	/*XXX doesn’t this leak? */
+}
+
+static void
+c_typeset_vardump(struct tbl *vp, uint32_t flag, int thing, bool pflag,
+    bool istset)
+{
+	struct tbl *tvp;
+	int any_set = 0;
+	char *s;
+
+	/*
+	 * See if the parameter is set (for arrays, if any
+	 * element is set).
+	 */
+	for (tvp = vp; tvp; tvp = tvp->u.array)
+		if (tvp->flag & ISSET) {
+			any_set = 1;
+			break;
+		}
+
+	/*
+	 * Check attributes - note that all array elements
+	 * have (should have?) the same attributes, so checking
+	 * the first is sufficient.
+	 *
+	 * Report an unset param only if the user has
+	 * explicitly given it some attribute (like export);
+	 * otherwise, after "echo $FOO", we would report FOO...
+	 */
+	if (!any_set && !(vp->flag & USERATTRIB))
+		return;
+	if (flag && (vp->flag & flag) == 0)
+		return;
+	if (!(vp->flag & ARRAY))
+		/* optimise later conditionals */
+		any_set = 0;
+	do {
+		/*
+		 * Ignore array elements that aren't set unless there
+		 * are no set elements, in which case the first is
+		 * reported on
+		 */
+		if (any_set && !(vp->flag & ISSET))
+			continue;
+		/* no arguments */
+		if (!thing && !flag) {
+			if (any_set == 1) {
+				shprintf("%s %s %s\n", Tset, "-A", vp->name);
+				any_set = 2;
+			}
+			/*
+			 * AT&T ksh prints things like export, integer,
+			 * leftadj, zerofill, etc., but POSIX says must
+			 * be suitable for re-entry...
+			 */
+			shprintf("%s %s", Ttypeset, "");
+			if (((vp->flag & (ARRAY | ASSOC)) == ASSOC))
+				shprintf("%s ", "-n");
+			if ((vp->flag & INTEGER))
+				shprintf("%s ", "-i");
+			if ((vp->flag & EXPORT))
+				shprintf("%s ", "-x");
+			if ((vp->flag & RDONLY))
+				shprintf("%s ", "-r");
+			if ((vp->flag & TRACE))
+				shprintf("%s ", "-t");
+			if ((vp->flag & LJUST))
+				shprintf("-L%d ", vp->u2.field);
+			if ((vp->flag & RJUST))
+				shprintf("-R%d ", vp->u2.field);
+			if ((vp->flag & ZEROFIL))
+				shprintf("%s ", "-Z");
+			if ((vp->flag & LCASEV))
+				shprintf("%s ", "-l");
+			if ((vp->flag & UCASEV_AL))
+				shprintf("%s ", "-u");
+			if ((vp->flag & INT_U))
+				shprintf("%s ", "-U");
+		} else if (pflag) {
+			shprintf("%s %s", istset ? Ttypeset :
+			    (flag & EXPORT) ? Texport : Treadonly, "");
+		}
+		if (any_set)
+			shprintf("%s[%lu]", vp->name, arrayindex(vp));
+		else
+			shf_puts(vp->name, shl_stdout);
+		if ((!thing && !flag && pflag) ||
+		    (thing == '-' && (vp->flag & ISSET))) {
+			s = str_val(vp);
+			shf_putc('=', shl_stdout);
+			/* AT&T ksh can't have justified integers... */
+			if ((vp->flag & (INTEGER | LJUST | RJUST)) == INTEGER)
+				shf_puts(s, shl_stdout);
+			else
+				print_value_quoted(shl_stdout, s);
+		}
+		shf_putc('\n', shl_stdout);
+
+		/*
+		 * Only report first 'element' of an array with
+		 * no set elements.
+		 */
+	} while (any_set && (vp = vp->u.array));
+}
+
 int
 c_alias(const char **wp)
 {
@@ -1053,7 +1061,7 @@ c_alias(const char **wp)
 				shf_puts(ap->name, shl_stdout);
 				if (prefix != '+') {
 					shf_putc('=', shl_stdout);
-					print_value_quoted(ap->val.s);
+					print_value_quoted(shl_stdout, ap->val.s);
 				}
 				shf_putc('\n', shl_stdout);
 			}
@@ -1078,7 +1086,7 @@ c_alias(const char **wp)
 				shf_puts(ap->name, shl_stdout);
 				if (prefix != '+') {
 					shf_putc('=', shl_stdout);
-					print_value_quoted(ap->val.s);
+					print_value_quoted(shl_stdout, ap->val.s);
 				}
 				shf_putc('\n', shl_stdout);
 			} else {
@@ -1282,7 +1290,7 @@ c_kill(const char **wp)
 	/* assume old style options if -digits or -UPPERCASE */
 	if ((p = wp[1]) && *p == '-' && (ksh_isdigit(p[1]) ||
 	    ksh_isupper(p[1]))) {
-		if (!(t = gettrap(p + 1, true))) {
+		if (!(t = gettrap(p + 1, false))) {
 			bi_errorf("bad signal '%s'", p + 1);
 			return (1);
 		}
@@ -1477,6 +1485,7 @@ c_getopts(const char **wp)
 	return (optc < 0 ? 1 : rv);
 }
 
+#ifndef MKSH_NO_CMDLINE_EDITING
 int
 c_bind(const char **wp)
 {
@@ -1535,6 +1544,7 @@ c_bind(const char **wp)
 
 	return (rv);
 }
+#endif
 
 int
 c_shift(const char **wp)
@@ -1759,11 +1769,11 @@ c_wait(const char **wp)
 	return (rv);
 }
 
+static char REPLY[] = "REPLY";
 int
 c_read(const char **wp)
 {
 #define is_ifsws(c) (ctype((c), C_IFS) && ctype((c), C_IFSWS))
-	static char REPLY[] = "REPLY";
 	int c, fd = 0, rv = 0, lastparm = 0;
 	bool savehist = false, intoarray = false, aschars = false;
 	bool rawmode = false, expanding = false;
@@ -1775,7 +1785,7 @@ c_read(const char **wp)
 	const char *ccp;
 	XString xs;
 	ptrdiff_t xsave = 0;
-	struct termios tios;
+	mksh_ttyst tios;
 	bool restore_tios = false;
 #if HAVE_SELECT
 	bool hastimeout = false;
@@ -1876,7 +1886,7 @@ c_read(const char **wp)
 
 #if HAVE_SELECT
 	if (hastimeout) {
-		gettimeofday(&tvlim, NULL);
+		mksh_TIME(tvlim);
 		timeradd(&tvlim, &tv, &tvlim);
 	}
 #endif
@@ -1888,7 +1898,7 @@ c_read(const char **wp)
 
 		FD_ZERO(&fdset);
 		FD_SET(fd, &fdset);
-		gettimeofday(&tv, NULL);
+		mksh_TIME(tv);
 		timersub(&tvlim, &tv, &tv);
 		if (tv.tv_sec < 0) {
 			/* timeout expired globally */
@@ -2021,7 +2031,7 @@ c_read(const char **wp)
 		vp = global(*wp);
 		if (vp->flag & RDONLY) {
  c_read_splitro:
-			bi_errorf("%s: %s", *wp, "is read only");
+			bi_errorf("read-only: %s", *wp);
  c_read_spliterr:
 			rv = 2;
 			afree(cp, ATEMP);
@@ -2152,7 +2162,7 @@ c_read(const char **wp)
 	afree(allocd, ATEMP);
 	Xfree(xs, xp);
 	if (restore_tios)
-		tcsetattr(fd, TCSADRAIN, &tios);
+		mksh_tcset(fd, &tios);
 	return (rv);
 #undef is_ifsws
 }
@@ -2225,7 +2235,7 @@ c_trap(const char **wp)
 		for (p = sigtraps, i = NSIG+1; --i >= 0; p++)
 			if (p->trap != NULL) {
 				shf_puts("trap -- ", shl_stdout);
-				print_value_quoted(p->trap);
+				print_value_quoted(shl_stdout, p->trap);
 				shprintf(" %s\n", p->name);
 			}
 		return (0);
@@ -2260,7 +2270,7 @@ c_exitreturn(const char **wp)
 	const char *arg;
 
 	if (ksh_getopt(wp, &builtin_opt, null) == '?')
-		return (1);
+		goto c_exitreturn_err;
 	arg = wp[builtin_opt.optind];
 
 	if (arg) {
@@ -2287,7 +2297,7 @@ c_exitreturn(const char **wp)
 	}
 
 	if (how == LEXIT && !really_exit && j_stopped_running()) {
-		really_exit = 1;
+		really_exit = true;
 		how = LSHELL;
 	}
 
@@ -2295,6 +2305,9 @@ c_exitreturn(const char **wp)
 	quitenv(NULL);
 	unwind(how);
 	/* NOTREACHED */
+
+ c_exitreturn_err:
+	return (1);
 }
 
 int
@@ -2305,19 +2318,19 @@ c_brkcont(const char **wp)
 	const char *arg;
 
 	if (ksh_getopt(wp, &builtin_opt, null) == '?')
-		return (1);
+		goto c_brkcont_err;
 	arg = wp[builtin_opt.optind];
 
 	if (!arg)
 		n = 1;
 	else if (!bi_getn(arg, &n))
-		return (1);
-	quit = n;
-	if (quit <= 0) {
+		goto c_brkcont_err;
+	if (n <= 0) {
 		/* AT&T ksh does this for non-interactive shells only - weird */
 		bi_errorf("%s: %s", arg, "bad value");
-		return (1);
+		goto c_brkcont_err;
 	}
+	quit = n;
 
 	/* Stop at E_NONE, E_PARSE, E_FUNC, or E_INCL */
 	for (ep = e; ep && !STOP_BRKCONT(ep->type); ep = ep->oenv)
@@ -2351,6 +2364,9 @@ c_brkcont(const char **wp)
 
 	unwind(*wp[0] == 'b' ? LBREAK : LCONTIN);
 	/* NOTREACHED */
+
+ c_brkcont_err:
+	return (1);
 }
 
 int
@@ -2434,8 +2450,7 @@ c_unset(const char **wp)
 			afree(cp, ATEMP);
 
 			if ((vp->flag&RDONLY)) {
-				warningf(true, "%s: %s", vp->name,
-				    "is read only");
+				warningf(true, "read-only: %s", vp->name);
 				rv = 1;
 			} else
 				unset(vp, optc);
@@ -2491,7 +2506,7 @@ timex(struct op *t, int f, volatile int *xerrok)
 	struct rusage ru0, ru1, cru0, cru1;
 	struct timeval usrtime, systime, tv0, tv1;
 
-	gettimeofday(&tv0, NULL);
+	mksh_TIME(tv0);
 	getrusage(RUSAGE_SELF, &ru0);
 	getrusage(RUSAGE_CHILDREN, &cru0);
 	if (t->left) {
@@ -2508,7 +2523,7 @@ timex(struct op *t, int f, volatile int *xerrok)
 		rv = execute(t->left, f | XTIME, xerrok);
 		if (t->left->type == TCOM)
 			tf |= t->left->str[0];
-		gettimeofday(&tv1, NULL);
+		mksh_TIME(tv1);
 		getrusage(RUSAGE_SELF, &ru1);
 		getrusage(RUSAGE_CHILDREN, &cru1);
 	} else
@@ -2930,7 +2945,11 @@ test_eval(Test_env *te, Test_op op, const char *opnd1, const char *opnd2,
 
 	/* -h or -L */
 	case TO_FILSYM:
+#ifdef MKSH__NO_SYMLINK
+		return (0);
+#else
 		return (lstat(opnd1, &b1) == 0 && S_ISLNK(b1.st_mode));
+#endif
 
 	/* -S */
 	case TO_FILSOCK:
@@ -2982,7 +3001,7 @@ test_eval(Test_env *te, Test_op op, const char *opnd1, const char *opnd2,
 
 	/* -s */
 	case TO_FILGZ:
-		return (stat(opnd1, &b1) == 0 && b1.st_size > 0L);
+		return (stat(opnd1, &b1) == 0 && (off_t)b1.st_size > (off_t)0);
 
 	/* -t */
 	case TO_FILTT:
@@ -3259,7 +3278,7 @@ ptest_error(Test_env *te, int ofs, const char *msg)
 struct limits {
 	const char *name;
 	int resource;		/* resource to get/set */
-	int factor;		/* multiply by to get rlim_{cur,max} values */
+	unsigned int factor;	/* multiply by to get rlim_{cur,max} values */
 	char option;
 };
 
@@ -3455,9 +3474,9 @@ set_ulimit(const struct limits *l, const char *v, int how)
 	if (strcmp(v, "unlimited") == 0)
 		val = (rlim_t)RLIM_INFINITY;
 	else {
-		mksh_ari_t rval;
+		mksh_uari_t rval;
 
-		if (!evaluate(v, &rval, KSH_RETURN_ERROR, false))
+		if (!evaluate(v, (mksh_ari_t *)&rval, KSH_RETURN_ERROR, false))
 			return (1);
 		/*
 		 * Avoid problems caused by typos that evaluate misses due
@@ -3507,7 +3526,7 @@ print_ulimit(const struct limits *l, int how)
 	if (val == (rlim_t)RLIM_INFINITY)
 		shf_puts("unlimited\n", shl_stdout);
 	else
-		shprintf("%ld\n", (long)(val / l->factor));
+		shprintf("%lu\n", (unsigned long)(val / l->factor));
 }
 #endif
 
@@ -3668,10 +3687,25 @@ c_sleep(const char **wp)
 		bi_errorf("%s: %s '%s'", Tsynerr, strerror(errno), wp[0]);
 	else {
 #ifndef MKSH_NOPROSPECTOFWORK
-		sigset_t omask;
+		sigset_t omask, bmask;
 
-		/* block SIGCHLD from interrupting us, though */
-		sigprocmask(SIG_BLOCK, &sm_sigchld, &omask);
+		/* block a number of signals from interrupting us, though */
+		(void)sigemptyset(&bmask);
+		(void)sigaddset(&bmask, SIGPIPE);
+		(void)sigaddset(&bmask, SIGCHLD);
+#ifdef SIGWINCH
+		(void)sigaddset(&bmask, SIGWINCH);
+#endif
+#ifdef SIGINFO
+		(void)sigaddset(&bmask, SIGINFO);
+#endif
+#ifdef SIGUSR1
+		(void)sigaddset(&bmask, SIGUSR1);
+#endif
+#ifdef SIGUSR2
+		(void)sigaddset(&bmask, SIGUSR2);
+#endif
+		sigprocmask(SIG_BLOCK, &bmask, &omask);
 #endif
 		if (select(1, NULL, NULL, NULL, &tv) == 0 || errno == EINTR)
 			/*
@@ -3682,20 +3716,10 @@ c_sleep(const char **wp)
 		else
 			bi_errorf("%s: %s", Tselect, strerror(errno));
 #ifndef MKSH_NOPROSPECTOFWORK
+		/* this will re-schedule signal delivery */
 		sigprocmask(SIG_SETMASK, &omask, NULL);
 #endif
 	}
 	return (rv);
 }
 #endif
-
-#if defined(ANDROID)
-static int
-c_android_lsmod(const char **wp MKSH_A_UNUSED)
-{
-	const char *cwp[3] = { "cat", "/proc/modules", NULL };
-
-	builtin_argv0 = cwp[0];
-	return (c_cat(cwp));
-}
-#endif
