@@ -1,10 +1,11 @@
 /*	$OpenBSD: main.c,v 1.47 2011/09/07 11:33:25 otto Exp $	*/
 /*	$OpenBSD: tty.c,v 1.9 2006/03/14 22:08:01 deraadt Exp $	*/
 /*	$OpenBSD: io.c,v 1.22 2006/03/17 16:30:13 millert Exp $	*/
-/*	$OpenBSD: table.c,v 1.13 2009/01/17 22:06:44 millert Exp $	*/
+/*	$OpenBSD: table.c,v 1.15 2012/02/19 07:52:30 otto Exp $	*/
 
 /*-
- * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
+ *		 2011, 2012
  *	Thorsten Glaser <tg@mirbsd.org>
  *
  * Provided that these terms and disclaimer and all copyright notices
@@ -33,7 +34,7 @@
 #include <locale.h>
 #endif
 
-__RCSID("$MirOS: src/bin/mksh/main.c,v 1.195.2.4 2011/11/08 22:07:21 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/main.c,v 1.222 2012/05/05 17:32:33 tg Exp $");
 
 extern char **environ;
 
@@ -45,6 +46,8 @@ extern char **environ;
 #define MKSH_DEFAULT_TMPDIR	"/tmp"
 #endif
 
+static uint8_t isuc(const char *);
+static int main_init(int, const char *[], Source **, struct block **);
 void chvt_reinit(void);
 static void reclaim(void);
 static void remove_temps(struct temp *);
@@ -94,7 +97,7 @@ static const char *restr_com[] = {
 	Ttypeset, "-r", "PATH", "ENV", "SHELL", NULL
 };
 
-static int initio_done;
+static bool initio_done;
 
 /* top-level parsing and execution environment */
 static struct env env;
@@ -107,9 +110,10 @@ rndsetup(void)
 	struct {
 		ALLOC_ITEM alloc_INT;
 		void *dataptr, *stkptr, *mallocptr;
+#if defined(__GLIBC__) && (__GLIBC__ >= 2)
 		sigjmp_buf jbuf;
+#endif
 		struct timeval tv;
-		struct timezone tz;
 	} *bufptr;
 	char *cp;
 
@@ -126,17 +130,19 @@ rndsetup(void)
 	bufptr->stkptr = &bufptr;
 	/* randomised malloc in BSD (and possibly others) */
 	bufptr->mallocptr = bufptr;
+#if defined(__GLIBC__) && (__GLIBC__ >= 2)
 	/* glibc pointer guard */
 	sigsetjmp(bufptr->jbuf, 1);
-	/* introduce variation */
-	gettimeofday(&bufptr->tv, &bufptr->tz);
+#endif
+	/* introduce variation (and yes, second arg MBZ for portability) */
+	mksh_TIME(bufptr->tv);
 
-	oaat1_init_impl(h);
+	NZATInit(h);
 	/* variation through pid, ppid, and the works */
-	oaat1_addmem_impl(h, &rndsetupstate, sizeof(rndsetupstate));
+	NZATUpdateMem(h, &rndsetupstate, sizeof(rndsetupstate));
 	/* some variation, some possibly entropy, depending on OE */
-	oaat1_addmem_impl(h, bufptr, sizeof(*bufptr));
-	oaat1_fini_impl(h);
+	NZATUpdateMem(h, bufptr, sizeof(*bufptr));
+	NZAATFinish(h);
 
 	afree(cp, APERM);
 	return ((mksh_uari_t)h);
@@ -155,8 +161,31 @@ static const char *empty_argv[] = {
 	"mksh", NULL
 };
 
-int
-main(int argc, const char *argv[])
+static uint8_t
+isuc(const char *cx) {
+	char *cp, *x;
+	uint8_t rv = 0;
+
+	if (!cx || !*cx)
+		return (0);
+
+	/* uppercase a string duplicate */
+	strdupx(x, cx, ATEMP);
+	cp = x;
+	while ((*cp = ksh_toupper(*cp)))
+		++cp;
+
+	/* check for UTF-8 */
+	if (strstr(x, "UTF-8") || strstr(x, "UTF8"))
+		rv = 1;
+
+	/* free copy and out */
+	afree(x, ATEMP);
+	return (rv);
+}
+
+static int
+main_init(int argc, const char *argv[], Source **sp, struct block **lp)
 {
 	int argi, i;
 	Source *s = NULL;
@@ -209,8 +238,8 @@ main(int argc, const char *argv[])
 
 	/* define built-in commands and see if we were called as one */
 	ktinit(APERM, &builtins,
-	    /* currently 50 builtins -> 80% of 64 (2^6) */
-	    6);
+	    /* currently up to 50 builtins */
+	    /* 80% of 64 = 2^6 */ 6, /* 66% of 128 = 2^7 */ 7);
 	for (i = 0; mkshbuiltins[i].name != NULL; i++)
 		if (!strcmp(ccp, builtin(mkshbuiltins[i].name,
 		    mkshbuiltins[i].func)))
@@ -238,10 +267,10 @@ main(int argc, const char *argv[])
 	coproc_init();
 
 	/* set up variable and command dictionaries */
-	ktinit(APERM, &taliases, 0);
-	ktinit(APERM, &aliases, 0);
+	ktinit(APERM, &taliases, 0, 0);
+	ktinit(APERM, &aliases, 0, 0);
 #ifndef MKSH_NOPWNAM
-	ktinit(APERM, &homedirs, 0);
+	ktinit(APERM, &homedirs, 0, 0);
 #endif
 
 	/* define shell keywords */
@@ -249,6 +278,13 @@ main(int argc, const char *argv[])
 
 	init_histvec();
 
+#ifdef TIOCGWINSZ
+	/* try to initialise tty size before importing environment */
+	tty_init(false, false);
+	change_winsz();
+	tty_close();
+#endif
+
 #ifdef _PATH_DEFPATH
 	def_path = _PATH_DEFPATH;
 #else
@@ -291,6 +327,7 @@ main(int argc, const char *argv[])
 	 */
 	Flag(FBRACEEXPAND) = 1;
 
+#ifndef MKSH_NO_CMDLINE_EDITING
 	/*
 	 * Set edit mode to emacs by default, may be overridden
 	 * by the environment or the user. Also, we want tab completion
@@ -299,6 +336,7 @@ main(int argc, const char *argv[])
 	change_flag(FEMACS, OF_SPECIAL, 1);
 #if !MKSH_S_NOVI
 	Flag(FVITABCOMPLETE) = 1;
+#endif
 #endif
 
 	/* import environment */
@@ -481,7 +519,9 @@ main(int argc, const char *argv[])
 			utf_flag = 0;
 #endif
 		}
+#ifndef MKSH_NO_CMDLINE_EDITING
 		x_init();
+#endif
 	}
 
 #ifdef SIGWINCH
@@ -503,8 +543,6 @@ main(int argc, const char *argv[])
 	}
 
 	/* divine the initial state of the utf8-mode Flag */
-#define isuc(x)	(((x) != NULL) && \
-	    (stristr((x), "UTF-8") || stristr((x), "utf8")))
 	ccp = null;
 	switch (utf_flag) {
 
@@ -546,7 +584,6 @@ main(int argc, const char *argv[])
 		UTFMODE = utf_flag;
 		break;
 	}
-#undef isuc
 
 	/* Disable during .profile/ENV reading */
 	restricted = Flag(FRESTRICTED);
@@ -594,15 +631,31 @@ main(int argc, const char *argv[])
 
 	alarm_init();
 
-	if (Flag(FAS_BUILTIN))
-		return (shcomexec(l->argv));
-
-	/* doesn't return */
-	shell(s, true);
-	/* NOTREACHED */
+	*sp = s;
+	*lp = l;
 	return (0);
 }
 
+/* this indirection barrier reduces stack usage during normal operation */
+
+int
+main(int argc, const char *argv[])
+{
+	int rv;
+	Source *s;
+	struct block *l;
+
+	if ((rv = main_init(argc, argv, &s, &l)) == 0) {
+		if (Flag(FAS_BUILTIN)) {
+			rv = shcomexec(l->argv);
+		} else {
+			shell(s, true);
+			/* NOTREACHED */
+		}
+	}
+	return (rv);
+}
+
 int
 include(const char *name, int argc, const char **argv, int intr_ok)
 {
@@ -624,8 +677,7 @@ include(const char *name, int argc, const char **argv, int intr_ok)
 		old_argc = 0;
 	}
 	newenv(E_INCL);
-	i = sigsetjmp(e->jbuf, 0);
-	if (i) {
+	if ((i = kshsetjmp(e->jbuf))) {
 		quitenv(s ? s->u.shf : NULL);
 		if (old_argv) {
 			e->loc->argv = old_argv;
@@ -687,62 +739,60 @@ command(const char *comm, int line)
  * run the commands from the input source, returning status.
  */
 int
-shell(Source * volatile s, volatile int toplevel)
+shell(Source * volatile s, volatile bool toplevel)
 {
 	struct op *t;
-	volatile int wastty = s->flags & SF_TTY;
-	volatile int attempts = 13;
-	volatile int interactive = Flag(FTALKING) && toplevel;
+	volatile bool wastty = tobool(s->flags & SF_TTY);
+	volatile uint8_t attempts = 13;
+	volatile bool interactive = Flag(FTALKING) && toplevel;
 	volatile bool sfirst = true;
 	Source *volatile old_source = source;
 	int i;
 
 	newenv(E_PARSE);
 	if (interactive)
-		really_exit = 0;
-	i = sigsetjmp(e->jbuf, 0);
-	if (i) {
-		switch (i) {
-		case LINTR:
-			/* we get here if SIGINT not caught or ignored */
-		case LERROR:
-		case LSHELL:
-			if (interactive) {
-				if (i == LINTR)
-					shellf("\n");
-				/*
-				 * Reset any eof that was read as part of a
-				 * multiline command.
-				 */
-				if (Flag(FIGNOREEOF) && s->type == SEOF &&
-				    wastty)
-					s->type = SSTDIN;
-				/*
-				 * Used by exit command to get back to
-				 * top level shell. Kind of strange since
-				 * interactive is set if we are reading from
-				 * a tty, but to have stopped jobs, one only
-				 * needs FMONITOR set (not FTALKING/SF_TTY)...
-				 */
-				/* toss any input we have so far */
-				s->start = s->str = null;
-				break;
-			}
-			/* FALLTHROUGH */
-		case LEXIT:
-		case LLEAVE:
-		case LRETURN:
-			source = old_source;
-			quitenv(NULL);
-			/* keep on going */
-			unwind(i);
-			/* NOTREACHED */
-		default:
-			source = old_source;
-			quitenv(NULL);
-			internal_errorf("%s %d", "shell", i);
-			/* NOTREACHED */
+		really_exit = false;
+	switch ((i = kshsetjmp(e->jbuf))) {
+	case 0:
+		break;
+	case LINTR:
+		/* we get here if SIGINT not caught or ignored */
+	case LERROR:
+	case LSHELL:
+		if (interactive) {
+			if (i == LINTR)
+				shellf("\n");
+			/*
+			 * Reset any eof that was read as part of a
+			 * multiline command.
+			 */
+			if (Flag(FIGNOREEOF) && s->type == SEOF && wastty)
+				s->type = SSTDIN;
+			/*
+			 * Used by exit command to get back to
+			 * top level shell. Kind of strange since
+			 * interactive is set if we are reading from
+			 * a tty, but to have stopped jobs, one only
+			 * needs FMONITOR set (not FTALKING/SF_TTY)...
+			 */
+			/* toss any input we have so far */
+			s->start = s->str = null;
+			break;
 		}
+		/* FALLTHROUGH */
+	case LEXIT:
+	case LLEAVE:
+	case LRETURN:
+		source = old_source;
+		quitenv(NULL);
+		/* keep on going */
+		unwind(i);
+		/* NOTREACHED */
+	default:
+		source = old_source;
+		quitenv(NULL);
+		internal_errorf("%s %d", "shell", i);
+		/* NOTREACHED */
 	}
 	while (/* CONSTCOND */ 1) {
 		if (trap)
@@ -766,7 +816,7 @@ shell(Source * volatile s, volatile int toplevel)
 				s->type = SSTDIN;
 			} else if (wastty && !really_exit &&
 			    j_stopped_running()) {
-				really_exit = 1;
+				really_exit = true;
 				s->type = SSTDIN;
 			} else {
 				/*
@@ -784,7 +834,7 @@ shell(Source * volatile s, volatile int toplevel)
 			exstat = execute(t, 0, NULL);
 
 		if (t != NULL && t->type != TEOF && interactive && really_exit)
-			really_exit = 0;
+			really_exit = false;
 
 		reclaim();
 	}
@@ -817,7 +867,7 @@ unwind(int i)
 		case E_INCL:
 		case E_LOOP:
 		case E_ERRH:
-			siglongjmp(e->jbuf, i);
+			kshlongjmp(e->jbuf, i);
 			/* NOTREACHED */
 		case E_NONE:
 			if (i == LINTR)
@@ -969,7 +1019,7 @@ remove_temps(struct temp *tp)
 {
 	for (; tp != NULL; tp = tp->next)
 		if (tp->pid == procpid)
-			unlink(tp->name);
+			unlink(tp->tffn);
 }
 
 /*
@@ -998,10 +1048,12 @@ tty_init(bool init_ttystate, bool need_tty)
 #endif
 	  if ((tfd = open("/dev/tty", O_RDWR, 0)) < 0) {
 		tty_devtty = false;
+#ifndef MKSH_DISABLE_TTY_WARNING
 		if (need_tty)
 			warningf(false, "%s: %s %s: %s",
 			    "No controlling tty", "open", "/dev/tty",
 			    strerror(errno));
+#endif
 	}
 	if (tfd < 0) {
 		do_close = false;
@@ -1010,8 +1062,10 @@ tty_init(bool init_ttystate, bool need_tty)
 		else if (isatty(2))
 			tfd = 2;
 		else {
+#ifndef MKSH_DISABLE_TTY_WARNING
 			if (need_tty)
 				warningf(false, "can't find tty fd");
+#endif
 			return;
 		}
 	}
@@ -1026,7 +1080,7 @@ tty_init(bool init_ttystate, bool need_tty)
 		close(tty_fd);
 		tty_fd = -1;
 	} else if (init_ttystate)
-		tcgetattr(tty_fd, &tty_state);
+		mksh_tcget(tty_fd, &tty_state);
 	if (do_close)
 		close(tfd);
 }
@@ -1217,7 +1271,13 @@ can_seek(int fd)
 	    SHF_UNBUF : 0);
 }
 
-struct shf shf_iob[3];
+#ifdef DF
+int shl_dbg_fd;
+#define NSHF_IOB 4
+#else
+#define NSHF_IOB 3
+#endif
+struct shf shf_iob[NSHF_IOB];
 
 void
 initio(void)
@@ -1225,9 +1285,23 @@ initio(void)
 	/* force buffer allocation */
 	shf_fdopen(1, SHF_WR, shl_stdout);
 	shf_fdopen(2, SHF_WR, shl_out);
-	/* force buffer allocation */
 	shf_fdopen(2, SHF_WR, shl_spare);
-	initio_done = 1;
+#ifdef DF
+	if ((shl_dbg_fd = open("/tmp/mksh-dbg.txt",
+	    O_WRONLY | O_APPEND | O_CREAT, 0600)) == -1)
+		errorf("cannot open debug output file");
+	if (shl_dbg_fd < FDBASE) {
+		int nfd;
+
+		nfd = fcntl(shl_dbg_fd, F_DUPFD, FDBASE);
+		close(shl_dbg_fd);
+		if ((shl_dbg_fd = nfd) == -1)
+			errorf("cannot dup debug output file");
+	}
+	shf_fdopen(shl_dbg_fd, SHF_WR, shl_dbg);
+	DF("=== open ===");
+#endif
+	initio_done = true;
 }
 
 /* A dup2() with error checking */
@@ -1433,42 +1507,61 @@ coproc_cleanup(int reuse)
 struct temp *
 maketemp(Area *ap, Temp_type type, struct temp **tlist)
 {
-	struct temp *tp;
+	char *cp;
 	size_t len;
-	int fd;
-	char *pathname;
+	int i;
+	struct temp *tp;
 	const char *dir;
+	struct stat sb;
 
 	dir = tmpdir ? tmpdir : MKSH_DEFAULT_TMPDIR;
-#if HAVE_MKSTEMP
-	len = strlen(dir) + 6 + 10 + 1;
-#else
-	pathname = tempnam(dir, "mksh.");
-	len = ((pathname == NULL) ? 0 : strlen(pathname)) + 1;
-#endif
-	/* reasonably sure that this will not overflow */
-	tp = alloc(sizeof(struct temp) + len, ap);
-	tp->name = (char *)&tp[1];
-#if !HAVE_MKSTEMP
-	if (pathname == NULL)
-		tp->name[0] = '\0';
-	else {
-		memcpy(tp->name, pathname, len);
-		free_ostempnam(pathname);
-	}
-#endif
-	pathname = tp->name;
-	tp->shf = NULL;
-	tp->type = type;
-#if HAVE_MKSTEMP
-	shf_snprintf(pathname, len, "%s%s", dir, "/mksh.XXXXXXXXXX");
-	if ((fd = mkstemp(pathname)) >= 0)
-#else
-	if (tp->name[0] && (fd = open(tp->name, O_CREAT | O_RDWR, 0600)) >= 0)
-#endif
-		tp->shf = shf_fdopen(fd, SHF_WR, NULL);
-	tp->pid = procpid;
+	/* add "/shXXXXXX.tmp" plus NUL */
+	len = strlen(dir);
+	checkoktoadd(len, offsetof(struct temp, tffn[0]) + 14);
+	tp = alloc(offsetof(struct temp, tffn[0]) + 14 + len, ap);
 
+	tp->shf = NULL;
+	tp->pid = procpid;
+	tp->type = type;
+
+	if (stat(dir, &sb) || !S_ISDIR(sb.st_mode)) {
+		tp->tffn[0] = '\0';
+		goto maketemp_out;
+	}
+
+	cp = (void *)tp;
+	cp += offsetof(struct temp, tffn[0]);
+	memcpy(cp, dir, len);
+	cp += len;
+	memcpy(cp, "/shXXXXXX.tmp", 14);
+	/* point to the first of six Xes */
+	cp += 3;
+	/* generate random part of filename */
+	len = -1;
+	do {
+		i = rndget() % 36;
+		cp[++len] = i < 26 ? 'a' + i : '0' + i - 26;
+	} while (len < 5);
+
+	/* cyclically attempt to open a temporary file */
+	while ((i = open(tp->tffn, O_CREAT | O_EXCL | O_RDWR, 0600)) == -1) {
+		if (errno != EEXIST)
+			goto maketemp_out;
+		/* count down from z to a then from 9 to 0 */
+		while (cp[len] == '0')
+			if (!len--)
+				goto maketemp_out;
+		if (cp[len] == 'a')
+			cp[len] = '9';
+		else
+			--cp[len];
+		/* do another cycle */
+	}
+
+	/* shf_fdopen cannot fail, so no fd leak */
+	tp->shf = shf_fdopen(i, SHF_WR, NULL);
+
+ maketemp_out:
 	tp->next = *tlist;
 	*tlist = tp;
 	return (tp);
@@ -1503,8 +1596,15 @@ tgrow(struct table *tp)
 	/* multiplication cannot overflow: alloc2 checked that */
 	memset(ntblp, 0, i * sizeof(struct tbl *));
 
-	/* table can get 80% full except when reaching its limit */
-	tp->nfree = (tp->tshift == 30) ? 0x3FFF0000UL : ((i * 4) / 5);
+	/* table can get very full when reaching its size limit */
+	tp->nfree = (tp->tshift == 30) ? 0x3FFF0000UL :
+	    /* but otherwise, only 80% (MKSH_SMALL) or 66% (normal) */
+#ifdef MKSH_SMALL
+	    ((i * 4) / 5)
+#else
+	    ((i * 2) / 3)
+#endif
+	    ;
 	tp->tbls = ntblp;
 	if (otblp == NULL)
 		return;
@@ -1534,7 +1634,7 @@ tgrow(struct table *tp)
 }
 
 void
-ktinit(Area *ap, struct table *tp, uint8_t initshift)
+ktinit_real(Area *ap, struct table *tp, uint8_t initshift)
 {
 	tp->areap = ap;
 	tp->tbls = NULL;
@@ -1662,3 +1762,54 @@ x_sigwinch(int sig MKSH_A_UNUSED)
 	got_winch = 1;
 }
 #endif
+
+#ifdef DF
+void
+DF(const char *fmt, ...)
+{
+	va_list args;
+	struct timeval tv;
+
+	mksh_lockfd(shl_dbg_fd);
+	mksh_TIME(tv);
+	shf_fprintf(shl_dbg, "[%d.%06d:%d] ", (int)tv.tv_sec, (int)tv.tv_usec,
+	    (int)getpid());
+	va_start(args, fmt);
+	shf_vfprintf(shl_dbg, fmt, args);
+	va_end(args);
+	shf_putc('\n', shl_dbg);
+	shf_flush(shl_dbg);
+	mksh_unlkfd(shl_dbg_fd);
+}
+#endif
+
+void
+x_mkraw(int fd, mksh_ttyst *ocb, bool forread)
+{
+	mksh_ttyst cb;
+
+	if (ocb)
+		mksh_tcget(fd, ocb);
+	else
+		ocb = &tty_state;
+
+	cb = *ocb;
+	if (forread) {
+		cb.c_lflag &= ~(ICANON) | ECHO;
+	} else {
+		cb.c_iflag &= ~(INLCR | ICRNL);
+		cb.c_lflag &= ~(ISIG | ICANON | ECHO);
+	}
+#if defined(VLNEXT) && defined(_POSIX_VDISABLE)
+	/* OSF/1 processes lnext when ~icanon */
+	cb.c_cc[VLNEXT] = _POSIX_VDISABLE;
+#endif
+	/* SunOS 4.1.x & OSF/1 processes discard(flush) when ~icanon */
+#if defined(VDISCARD) && defined(_POSIX_VDISABLE)
+	cb.c_cc[VDISCARD] = _POSIX_VDISABLE;
+#endif
+	cb.c_cc[VTIME] = 0;
+	cb.c_cc[VMIN] = 1;
+
+	mksh_tcset(fd, &cb);
+}
