@@ -79,7 +79,12 @@
 #if HAVE_STRINGS_H
 #include <strings.h>
 #endif
+#if HAVE_TERMIOS_H
 #include <termios.h>
+#else
+/* shudder… */
+#include <termio.h>
+#endif
 #include <time.h>
 #if HAVE_ULIMIT_H
 #include <ulimit.h>
@@ -152,9 +157,9 @@
 #endif
 
 #ifdef EXTERN
-__RCSID("$MirOS: src/bin/mksh/sh.h,v 1.549 2012/04/14 16:07:48 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/sh.h,v 1.560 2012/05/09 23:20:58 tg Exp $");
 #endif
-#define MKSH_VERSION "R40 2012/04/14"
+#define MKSH_VERSION "R40 2012/05/09"
 
 /* arithmetic types: C implementation */
 #if !HAVE_CAN_INTTYPES
@@ -284,7 +289,7 @@ struct rusage {
 #define ksh_isdash(s)	(((s) != NULL) && ((s)[0] == '-') && ((s)[1] == '\0'))
 #define ksh_isspace(c)	((((c) >= 0x09) && ((c) <= 0x0D)) || ((c) == 0x20))
 
-#ifdef NO_PATH_MAX
+#ifdef MKSH__NO_PATH_MAX
 #undef PATH_MAX
 #else
 #ifndef PATH_MAX
@@ -349,6 +354,37 @@ extern int __cdecl seteuid(uid_t);
 extern int __cdecl setegid(gid_t);
 #endif
 
+#if defined(__COHERENT__)
+#ifndef O_ACCMODE
+/* this need not work everywhere, take care */
+#define O_ACCMODE	(O_RDONLY | O_WRONLY | O_RDWR)
+#endif
+#define mksh_TIME(tv) do {		\
+	(tv).tv_usec = 0;		\
+	(tv).tv_sec = time(NULL);	\
+} while (/* CONSTCOND */ 0)
+#else
+#define mksh_TIME(tv) gettimeofday(&(tv), NULL)
+#endif
+
+#ifdef MKSH__NO_SYMLINK
+#undef S_ISLNK
+#define S_ISLNK(m)	(/* CONSTCOND */ 0)
+#define mksh_lstat	stat
+#else
+#define mksh_lstat	lstat
+#endif
+
+#if HAVE_TERMIOS_H
+#define mksh_ttyst	struct termios
+#define mksh_tcget(fd,st) tcgetattr((fd), (st))
+#define mksh_tcset(fd,st) tcsetattr((fd), TCSADRAIN, (st))
+#else
+#define mksh_ttyst	struct termio
+#define mksh_tcget(fd,st) ioctl((fd), TCGETA, (st))
+#define mksh_tcset(fd,st) ioctl((fd), TCSETAW, (st))
+#endif
+
 /* remove redundancies */
 
 #if defined(MirBSD) && (MirBSD >= 0x08A8) && !defined(MKSH_OPTSTATIC)
@@ -535,7 +571,7 @@ im_sorry_dave(void)
 #define free_ossetmode(p)	free(p)
 #endif
 
-#ifdef NO_PATH_MAX
+#ifdef MKSH__NO_PATH_MAX
 /* GNU libc: get_current_dir_name(3) -> free(3) */
 #define free_gnu_gcdn(p)	free(p)
 #endif
@@ -694,6 +730,12 @@ EXTERN const char Tpunalias[] E_INIT("+unalias");
 #define Tunalias	(Tpunalias + 1)		/* "unalias" */
 EXTERN const char Tsgset[] E_INIT("*=set");
 #define Tset		(Tsgset + 2)		/* "set" */
+EXTERN const char Tsgunset[] E_INIT("*=unset");
+#define Tunset		(Tsgunset + 2)		/* "unset" */
+EXTERN const char Tsgexport[] E_INIT("*=export");
+#define Texport		(Tsgexport + 2)		/* "export" */
+EXTERN const char Tsgreadonly[] E_INIT("*=readonly");
+#define Treadonly	(Tsgreadonly + 2)	/* "readonly" */
 EXTERN const char Tgbuiltin[] E_INIT("=builtin");
 #define Tbuiltin	(Tgbuiltin + 1)		/* "builtin" */
 EXTERN const char T_function[] E_INIT(" function");
@@ -792,7 +834,7 @@ EXTERN unsigned int ksh_tmout;
 EXTERN enum tmout_enum ksh_tmout_state E_INIT(TMOUT_EXECUTING);
 
 /* For "You have stopped jobs" message */
-EXTERN int really_exit;
+EXTERN bool really_exit;
 
 /*
  * fast character classes
@@ -1491,7 +1533,7 @@ EXTERN struct timeval j_usrtime, j_systime;
 } while (/* CONSTCOND */ 0)
 
 
-/* NZAT/NZAAT hashes based on Bob Jenkins' one-at-a-time hash */
+/* NZAAT hash based on Bob Jenkins' one-at-a-time hash */
 
 /* From: src/kern/include/nzat.h,v 1.2 2011/07/18 00:35:40 tg Exp $ */
 
@@ -1524,15 +1566,6 @@ EXTERN struct timeval j_usrtime, j_systime;
 		NZATUpdateByte((h), NZATUpdateString_c);	\
 } while (/* CONSTCOND */ 0)
 
-/* not zero after termination */
-#define NZATFinish(h) do {					\
-	if ((h) == 0)						\
-		++(h);						\
-	else							\
-		NZAATFinish(h);					\
-} while (/* CONSTCOND */ 0)
-
-/* NULs zählen an allen Teilen */
 #define NZAATFinish(h) do {					\
 	(h) += (h) << 10;					\
 	(h) ^= (h) >> 6;					\
@@ -1552,14 +1585,16 @@ void *aresize(void *, size_t, Area *);
 void *aresize2(void *, size_t, size_t, Area *);
 void afree(void *, Area *);	/* can take NULL */
 /* edit.c */
+#ifndef MKSH_NO_CMDLINE_EDITING
 #ifndef MKSH_SMALL
 int x_bind(const char *, const char *, bool, bool);
 #else
 int x_bind(const char *, const char *, bool);
 #endif
 void x_init(void);
-void x_mkraw(int, struct termios *, bool);
 int x_read(char *, size_t);
+#endif
+void x_mkraw(int, mksh_ttyst *, bool);
 /* eval.c */
 char *substitute(const char *, int);
 char **eval(const char **, int);
@@ -1613,7 +1648,9 @@ int c_fgbg(const char **);
 int c_kill(const char **);
 void getopts_reset(int);
 int c_getopts(const char **);
+#ifndef MKSH_NO_CMDLINE_EDITING
 int c_bind(const char **);
+#endif
 int c_shift(const char **);
 int c_umask(const char **);
 int c_dot(const char **);
@@ -1655,7 +1692,7 @@ void sethistsize(mksh_ari_t);
 #if HAVE_PERSISTENT_HISTORY
 void sethistfile(const char *);
 #endif
-#if !MKSH_S_NOVI
+#if !defined(MKSH_NO_CMDLINE_EDITING) && !MKSH_S_NOVI
 char **histpos(void);
 int histnum(int);
 #endif
@@ -1711,7 +1748,7 @@ int promptlen(const char *);
 /* main.c */
 int include(const char *, int, const char **, int);
 int command(const char *, int);
-int shell(Source *volatile, int volatile);
+int shell(Source * volatile, volatile bool);
 void unwind(int) MKSH_A_NORETURN;
 void newenv(int);
 void quitenv(struct shf *);
@@ -1781,7 +1818,6 @@ char *getoptions(void);
 void change_flag(enum sh_flag, int, unsigned int);
 int parse_args(const char **, int, bool *);
 int getn(const char *, int *);
-int bi_getn(const char *, int *);
 int gmatchx(const char *, const char *, bool);
 int has_globbing(const char *, const char *);
 int xstrcmp(const void *, const void *);
@@ -1865,6 +1901,7 @@ void fpFUNCTf(struct shf *, int, bool, const char *, struct op *);
 void newblock(void);
 void popblock(void);
 void initvar(void);
+struct block *varsearch(struct block *, struct tbl **, const char *, uint32_t);
 struct tbl *global(const char *);
 struct tbl *local(const char *, bool);
 char *str_val(struct tbl *);
@@ -1944,7 +1981,7 @@ int test_parse(Test_env *);
 
 EXTERN int tty_fd E_INIT(-1);	/* dup'd tty file descriptor */
 EXTERN bool tty_devtty;		/* true if tty_fd is from /dev/tty */
-EXTERN struct termios tty_state;	/* saved tty state */
+EXTERN mksh_ttyst tty_state;	/* saved tty state */
 
 extern void tty_init(bool, bool);
 extern void tty_close(void);
