@@ -27,7 +27,7 @@
 #include <sys/sysctl.h>
 #endif
 
-__RCSID("$MirOS: src/bin/mksh/var.c,v 1.146 2012/04/14 14:35:13 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/var.c,v 1.149 2012/05/09 23:21:00 tg Exp $");
 
 /*-
  * Variables
@@ -141,8 +141,8 @@ initvar(void)
 	}
 }
 
-/* common code for several functions below */
-static struct block *
+/* common code for several functions below and c_typeset() */
+struct block *
 varsearch(struct block *l, struct tbl **vpp, const char *vn, uint32_t h)
 {
 	register struct tbl *vp;
@@ -1088,7 +1088,7 @@ getspec(struct tbl *vp)
 		if (vp->flag & ISSET) {
 			struct timeval tv;
 
-			gettimeofday(&tv, NULL);
+			mksh_TIME(tv);
 			i = tv.tv_sec - seconds;
 		} else
 			return;
@@ -1221,7 +1221,7 @@ setspec(struct tbl *vp)
 		{
 			struct timeval tv;
 
-			gettimeofday(&tv, NULL);
+			mksh_TIME(tv);
 			seconds = tv.tv_sec - i;
 		}
 		break;
@@ -1471,7 +1471,7 @@ hash(const void *s)
 
 	NZATInit(h);
 	NZATUpdateString(h, s);
-	NZATFinish(h);
+	NZAATFinish(h);
 	return (h);
 }
 
